@@ -26,6 +26,7 @@
 #include "polymake/hash_map"
 #include "polymake/vector"
 #include "polymake/list"
+#include "polymake/optional"
 #include <algorithm>
 #include <sstream>
 
@@ -197,73 +198,61 @@ struct permutation_map<Value, operations::cmp, std::true_type, typename std::ena
 };
 
 template <typename Input1, typename Input2, typename Output, typename Comparator, typename ExpectDuplicates>
-void find_permutation_impl(Input1&& src1, Input2&& src2, Output&& dst, const Comparator& comparator, ExpectDuplicates)
+bool find_permutation_impl(Input1&& src1, Input2&& src2, Output&& dst, const Comparator& comparator, ExpectDuplicates)
 {
    typename permutation_map<typename iterator_traits<Input1>::value_type, Comparator, ExpectDuplicates>::type index_map;
 
-   for (int i=0; !src1.at_end(); ++src1, ++i)
+   for (int i = 0; !src1.at_end(); ++src1, ++i)
       index_map.emplace(*src1, i);
 
    for (; !src2.at_end(); ++src2, ++dst) {
-      const auto where=index_map.find(*src2);
-      if (where == index_map.end()) {
-         std::string reason;
-         if (index_map.empty()) {
-            reason="not a permutation: first sequence is shorter";
-         } else {
-            std::ostringstream os;
-            wrap(os) << "not a permutation: no match for <" << *src2 << ">";
-            reason=os.str();
-         }
-         throw no_match(reason);
-      }
-      *dst=where->second;
+      const auto where = index_map.find(*src2);
+      if (where == index_map.end())
+         return false;
+      *dst = where->second;
       index_map.erase(where);
    }
 
-   if (!index_map.empty())
-      throw no_match("not a permutation: second sequence is shorter");
+   return index_map.empty();
 }
 
-template <typename Container1, typename Container2, typename Comparator=polymake::operations::cmp>
-Array<int> find_permutation(const Container1& c1, const Container2& c2, const Comparator& comparator=Comparator())
+template <typename Container1, typename Container2, typename Comparator = polymake::operations::cmp>
+optional<Array<int>>
+find_permutation(const Container1& c1, const Container2& c2, const Comparator& comparator = Comparator())
 {
    Array<int> perm(c1.size());
-   find_permutation_impl(entire(c1), entire(c2), perm.begin(), comparator, std::false_type());
-   return perm;
+   if (find_permutation_impl(entire(c1), entire(c2), perm.begin(), comparator, std::false_type()))
+      return make_optional(std::move(perm));
+   else
+      return nullopt;
 }
 
-template <typename Container1, typename Container2, typename Comparator=polymake::operations::cmp>
-Array<int> find_permutation_with_duplicates(const Container1& c1, const Container2& c2, const Comparator& comparator=Comparator())
+template <typename Container1, typename Container2, typename Comparator = polymake::operations::cmp>
+optional<Array<int>>
+find_permutation_with_duplicates(const Container1& c1, const Container2& c2, const Comparator& comparator = Comparator())
 {
    Array<int> perm(c1.size());
-   find_permutation_impl(entire(c1), entire(c2), perm.begin(), comparator, std::true_type());
-   return perm;
+   if (find_permutation_impl(entire(c1), entire(c2), perm.begin(), comparator, std::true_type()))
+      return make_optional(std::move(perm));
+   else
+      return nullopt;
+}
+
+template <typename Container1, typename Container2, typename Comparator = polymake::operations::cmp>
+bool are_permuted(const Container1& c1, const Container2& c2, const Comparator& comparator = Comparator())
+{
+   return !!find_permutation(c1, c2, comparator);
 }
 
 template <typename Container1, typename Container2, typename Comparator=polymake::operations::cmp>
-bool are_permuted(const Container1& c1, const Container2& c2, const Comparator& comparator=Comparator())
+bool are_permuted_with_duplicates(const Container1& c1, const Container2& c2, const Comparator& comparator = Comparator())
 {
-   try {
-      find_permutation(c1, c2, comparator);
-      return true;
-   } catch (const no_match&) {}
-   return false;
-}
-
-template <typename Container1, typename Container2, typename Comparator=polymake::operations::cmp>
-bool are_permuted_with_duplicates(const Container1& c1, const Container2& c2, const Comparator& comparator=Comparator())
-{
-   try {
-      find_permutation_with_duplicates(c1, c2, comparator);
-      return true;
-   } catch (const no_match&) {}
-   return false;
+   return !!find_permutation_with_duplicates(c1, c2, comparator);
 }
 
 template <typename Container, typename Permutation>
-typename std::enable_if<std::is_same<typename object_traits<Container>::generic_tag, is_container>::value,
-                        typename object_traits<Container>::persistent_type>::type
+std::enable_if_t<std::is_same<typename object_traits<Container>::generic_tag, is_container>::value,
+                 typename object_traits<Container>::persistent_type>
 permuted(const Container& c, const Permutation& perm)
 {
    if (POLYMAKE_DEBUG) {
@@ -276,8 +265,8 @@ permuted(const Container& c, const Permutation& perm)
 }
 
 template <typename Container, typename Permutation>
-typename std::enable_if<std::is_same<typename object_traits<Container>::generic_tag, is_container>::value,
-                        typename object_traits<Container>::persistent_type>::type
+std::enable_if_t<std::is_same<typename object_traits<Container>::generic_tag, is_container>::value,
+                 typename object_traits<Container>::persistent_type>
 permuted_inv(const Container& c, const Permutation& perm)
 {
    if (POLYMAKE_DEBUG) {
@@ -304,11 +293,11 @@ class permutation_iterator;
 
 class permutation_iterator_base {
 public:
-   typedef forward_iterator_tag iterator_category;
-   typedef std::vector<int> value_type;
-   typedef const value_type* pointer;
-   typedef const value_type& reference;
-   typedef ptrdiff_t difference_type;
+   using iterator_category = forward_iterator_tag;
+   using value_type = Array<int>;
+   using reference = value_type;
+   using pointer = const value_type*;
+   using difference_type = ptrdiff_t;
 protected:
    value_type perm;
 
@@ -327,8 +316,8 @@ class permutation_iterator<permutations_heap>
    : public permutation_iterator_base {
    friend class AllPermutations<permutations_heap>;
 public:
-   typedef permutation_iterator iterator;
-   typedef iterator const_iterator;
+   using iterator = permutation_iterator;
+   using const_iterator = iterator;
 
 protected:
    std::vector<int> cnt;
@@ -349,18 +338,18 @@ public:
       : permutation_iterator_base(n_arg)
       , cnt(size_t(n_arg), 0)
       , n(n_arg)
-      , pos(n_arg>1) {}
+      , pos(n_arg > 1) {}
 
    iterator& operator++ ()
    {
       do {
-         if (cnt[pos]<pos) {
+         if (cnt[pos] < pos) {
             std::swap(perm[pos], perm[pos%2 * cnt[pos]]);
             ++cnt[pos];
-            pos=1;
+            pos = 1;
             break;
          }
-         cnt[pos]=0;
+         cnt[pos] = 0;
       } while (++pos < n);
       return *this;
    }
@@ -370,7 +359,7 @@ public:
 
    bool operator== (const iterator& it) const
    {
-      return n==it.n && pos==it.pos && cnt==it.cnt;
+      return n == it.n && pos == it.pos && cnt == it.cnt;
    }
    bool operator!= (const iterator& it) const
    {
@@ -380,7 +369,7 @@ public:
    void rewind()
    {
       reset(n);
-      pos=1;
+      pos = 1;
       fill_range(entire(cnt), 0);
    }
 };
@@ -400,31 +389,31 @@ protected:
    void incr()
    {
       if (next==0) {
-         next=-1; return;
+         next = -1; return;
       }
-      int lim=n, offset=0;
-      std::vector<int>::iterator pos=move.begin();
+      int lim = n, offset = 0;
+      auto pos = move.begin();
       for (;;) {
-         int m=*pos;
-         if (m>0) {
-            if (m<lim) {
-               next=m+offset;
+         const int m = *pos;
+         if (m > 0) {
+            if (m < lim) {
+               next = m + offset;
                ++*pos;
                break;
             }
          } else {
-            if (m<-1) {
-               next=offset-m-1;
+            if (m < -1) {
+               next = offset - m - 1;
                ++*pos;
                break;
             }
             ++offset;
          }
          if (--lim > 1) {
-            *pos=-m;
+            *pos = -m;
             ++pos;
          } else {
-            next=0;
+            next = 0;
             break;
          }
       }
@@ -439,8 +428,8 @@ public:
       , n(n_arg)
       , next(n-1)
    {
-      if (next>0) {
-         next=1; move[0]=2;
+      if (next > 0) {
+         next = 1; move[0] = 2;
       }
    }
 
@@ -469,9 +458,9 @@ public:
    {
       reset(n);
       fill_range(entire(move), 1);
-      next=n-1;
-      if (next>1) {
-         next=1; move[0]=2;
+      next = n - 1;
+      if (next > 1) {
+         next = 1; move[0] = 2;
       }
    }
 };
